﻿Imports System.Collections.ObjectModel
Imports System.Globalization
Imports System.ServiceModel
Imports Microsoft.Xrm.Sdk

Namespace NamespaceName
    Partial Public MustInherit Class BasePluginClassName
        Implements IPlugin
        Protected Class LocalPluginContext(Of T As Entity)
            Implements IDisposable
            Friend Property CrmContext As Microsoft.Xrm.Sdk.Client.OrganizationServiceContext

            Friend Property ServiceProvider As IServiceProvider

            Friend Property ServiceFactory As IOrganizationServiceFactory

            Friend Property OrganizationService As IOrganizationService

            Friend Property PluginExecutionContext As IPluginExecutionContext

            Friend Property TracingService As ITracingService

            Friend ReadOnly Property Stage As eStage
                Get
                    Return CType(Me.PluginExecutionContext.Stage, eStage)
                End Get
            End Property
            Friend ReadOnly Property Depth As Integer
                Get
                    Return Me.PluginExecutionContext.Depth
                End Get
            End Property
            Friend ReadOnly Property MessageName As String
                Get
                    Return Me.PluginExecutionContext.MessageName
                End Get
            End Property
            Friend Sub New(serviceProvider As IServiceProvider)
                If serviceProvider Is Nothing Then
                    Throw New ArgumentNullException("serviceProvider")
                End If

                ' Obtain the tracing service from the service provider.
                Me.TracingService = DirectCast(serviceProvider.GetService(GetType(ITracingService)), ITracingService)

                ' Obtain the execution context service from the service provider.
                Me.PluginExecutionContext = DirectCast(serviceProvider.GetService(GetType(IPluginExecutionContext)), IPluginExecutionContext)

                ' Obtain the Organization Service factory service from the service provider
                Me.ServiceFactory = DirectCast(serviceProvider.GetService(GetType(IOrganizationServiceFactory)), IOrganizationServiceFactory)

                ' Use the factory to generate the Organization Service.
                Me.OrganizationService = Me.ServiceFactory.CreateOrganizationService(Me.PluginExecutionContext.UserId)

                ' Generate the CrmContext to use with LINQ etc
                Me.CrmContext = New Microsoft.Xrm.Sdk.Client.OrganizationServiceContext(Me.OrganizationService)
            End Sub

            Friend Sub Trace(message As String)
                If String.IsNullOrWhiteSpace(message) OrElse Me.TracingService Is Nothing Then
                    Return
                End If

                If Me.PluginExecutionContext Is Nothing Then
                    Me.TracingService.Trace(message)
                Else
                    Me.TracingService.Trace("{0}, Correlation Id: {1}, Initiating User: {2}", message, Me.PluginExecutionContext.CorrelationId, Me.PluginExecutionContext.InitiatingUserId)
                End If
            End Sub

            Public Sub Dispose() Implements IDisposable.Dispose
                If Me.CrmContext IsNot Nothing Then
                    Me.CrmContext.Dispose()
                End If
            End Sub
            ''' <summary>
            ''' Returns the first registered 'Pre' image for the pipeline execution
            ''' </summary>
            Friend ReadOnly Property PreImage As T
                Get
                    If Me.PluginExecutionContext.PreEntityImages.Any() Then
                        Return GetEntityAsType(Me.PluginExecutionContext.PreEntityImages(Me.PluginExecutionContext.PreEntityImages.FirstOrDefault().Key))
                    End If
                    Return Nothing
                End Get
            End Property
            ''' <summary>
            ''' Returns the first registered 'Post' image for the pipeline execution
            ''' </summary>
            Friend ReadOnly Property PostImage As T
                Get
                    If Me.PluginExecutionContext.PostEntityImages.Any() Then
                        Return GetEntityAsType(Me.PluginExecutionContext.PostEntityImages(Me.PluginExecutionContext.PostEntityImages.FirstOrDefault().Key))
                    End If
                    Return Nothing
                End Get
            End Property
            ''' <summary>
            ''' Returns the 'Target' of the message if available
            ''' This is an 'Entity' instead of the specified type in order to retain the same instance of the 'Entity' object. This allows for updates to the target in a 'Pre' stage that
            ''' will get persisted during the transaction.
            ''' </summary>
            Friend ReadOnly Property TargetEntity As Entity
                Get
                    If Me.PluginExecutionContext.InputParameters.Contains("Target") Then
                        Return TryCast(Me.PluginExecutionContext.InputParameters("Target"), Entity)
                    End If
                    Return Nothing
                End Get
            End Property
            ''' <summary>
            ''' Returns the 'Target' of the message as an EntityReference if available
            ''' </summary>
            Friend ReadOnly Property TargetEntityReference As EntityReference
                Get
                    If Me.PluginExecutionContext.InputParameters.Contains("Target") Then
                        Return TryCast(Me.PluginExecutionContext.InputParameters("Target"), EntityReference)
                    End If
                    Return Nothing
                End Get
            End Property
            Private Function GetEntityAsType(entity As Entity) As T
                If GetType(T) = entity.GetType() Then
                    Return TryCast(entity, T)
                Else
                    Return entity.ToEntity(Of T)()
                End If
            End Function
        End Class
        Protected Enum eStage
            PreValidation = 10
            PreOperation = 20
            PostOperation = 40
        End Enum
        Protected Class PluginEvent
            ''' <summary>
            ''' Execution pipeline stage that the plugin should be registered against.
            ''' </summary>
            Public Property Stage As eStage

            ''' <summary>
            ''' Logical name of the entity that the plugin should be registered against. Leave 'null' to register against all entities.
            ''' </summary>
            Public Property EntityName As String

            ''' <summary>
            ''' Name of the message that the plugin should be triggered off of.
            ''' </summary>
            Public Property MessageName As String

            ''' <summary>
            ''' Method that should be executed when the conditions of the Plugin Event have been met.
            ''' </summary>
            Public Property PluginAction As Action(Of IServiceProvider)
        End Class

        Private _registeredEvents As Collection(Of PluginEvent)

        ''' <summary>
        ''' Gets the List of events that the plug-in should fire for. Each List
        ''' </summary>
        Protected ReadOnly Property RegisteredEvents As Collection(Of PluginEvent)
            Get
                If Me._registeredEvents Is Nothing Then
                    Me._registeredEvents = New Collection(Of PluginEvent)()
                End If
                Return Me._registeredEvents
            End Get
        End Property

        ''' <summary>
        ''' Initializes a new instance of the BasePluginClassName class.
        ''' </summary>
        Friend Sub New(unsecureConfig As String, secureConfig As String)
            Me.UnsecureConfig = unsecureConfig
            Me.SecureConfig = secureConfig
        End Sub
        ''' <summary>
        ''' Un secure configuration specified during the registration of the plugin step
        ''' </summary>
        Public Property UnsecureConfig As String

        ''' <summary>
        ''' Secure configuration specified during the registration of the plugin step
        ''' </summary>
        Public Property SecureConfig As String

        ''' <summary>
        ''' Executes the plug-in.
        ''' </summary>
        ''' <param name="serviceProvider">The service provider.</param>
        ''' <remarks>
        ''' For improved performance, Microsoft Dynamics CRM caches plug-in instances. 
        ''' The plug-in's Execute method should be written to be stateless as the constructor 
        ''' is not called for every invocation of the plug-in. Also, multiple system threads 
        ''' could execute the plug-in at the same time. All per invocation state information 
        ''' is stored in the context. This means that you should not use global variables in plug-ins.
        ''' </remarks>
        Public Sub Execute(serviceProvider As IServiceProvider) Implements IPlugin.Execute
            If serviceProvider Is Nothing Then
                Throw New ArgumentNullException("serviceProvider")
            End If

            Dim tracingService = DirectCast(serviceProvider.GetService(GetType(ITracingService)), ITracingService)
            Dim pluginContext = DirectCast(serviceProvider.GetService(GetType(IPluginExecutionContext)), IPluginExecutionContext)

            tracingService.Trace(String.Format(CultureInfo.InvariantCulture, "Entered {0}.Execute()", Me.GetType().ToString()))
            Try
                ' Iterate over all of the expected registered events to ensure that the plugin
                ' has been invoked by an expected event
                Dim entityActions = (From a In Me.RegisteredEvents Where (CInt(a.Stage) = pluginContext.Stage AndAlso (If(String.IsNullOrWhiteSpace(a.MessageName), True, a.MessageName.ToLowerInvariant() = pluginContext.MessageName.ToLowerInvariant())) AndAlso (If(String.IsNullOrWhiteSpace(a.EntityName), True, a.EntityName.ToLowerInvariant() = pluginContext.PrimaryEntityName.ToLowerInvariant()))) Select a.PluginAction)

                If entityActions.Any() Then
                    For Each entityAction As Action(Of IServiceProvider) In entityActions
                        tracingService.Trace(String.Format(CultureInfo.InvariantCulture, "{0} is firing for Entity: {1}, Message: {2}, Method: {3}", Me.GetType().ToString(), pluginContext.PrimaryEntityName, pluginContext.MessageName, entityAction.Method.Name))
                        entityAction.Invoke(serviceProvider)
                    Next
                End If
            Catch ex As Exception
                tracingService.Trace(String.Format(CultureInfo.InvariantCulture, "Exception: {0}", ex.ToString()))
                Throw
            Finally
                tracingService.Trace(String.Format(CultureInfo.InvariantCulture, "Exiting {0}.Execute()", Me.GetType().ToString()))
            End Try
        End Sub
    End Class
    Public Structure EntityNames
        Public Shared ReadOnly EntityNamesPlaceHolder As String
    End Structure
    Public Structure MessageNames
        Public Shared ReadOnly MessageNamesPlaceHolder As String
    End Structure
End Namespace
